from .base import *
from .normalize import Normalize
from .min_max import MinMax
from .function import Lambda
from .one_hot import OneHot
from .identical import Identical


__all__ = ["Processor", "processor_dict"]
